@@ -1,4 +1,7 @@
-/* Copyright Joyent, Inc. and other Node contributors. All rights reserved.
+/* Based on src/http/ngx_http_parse.c from NGINX copyright Igor Sysoev
+ *
+ * Additional changes are licensed under the same terms as NGINX and
+ * copyright Joyent, Inc. and other Node contributors. All rights reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to
@@ -97,6 +100,7 @@ static const char *method_strings[] =
   , "NOTIFY"
   , "SUBSCRIBE"
   , "UNSUBSCRIBE"
+  , "PATCH"
   };
 
 
@@ -186,7 +190,7 @@ static const uint8_t normal_url_char[256] = {
 /* 112  p   113  q   114  r   115  s   116  t   117  u   118  v   119  w  */
         1,       1,       1,       1,       1,       1,       1,       1,
 /* 120  x   121  y   122  z   123  {   124  |   125  }   126  ~   127 del */
-        1,       1,       1,       1,       1,       1,       1,       0 };
+        1,       1,       1,       1,       1,       1,       1,       0, };
 
 
 enum state
@@ -237,6 +241,7 @@ enum state
   , s_header_field
   , s_header_value_start
   , s_header_value
+  , s_header_value_lws
 
   , s_header_almost_done
 
@@ -290,20 +295,24 @@ enum header_states
   };
 
 
-enum flags
-  { F_CHUNKED               = 1 << 0
-  , F_CONNECTION_KEEP_ALIVE = 1 << 1
-  , F_CONNECTION_CLOSE      = 1 << 2
-  , F_TRAILING              = 1 << 3
-  , F_UPGRADE               = 1 << 4
-  , F_SKIPBODY              = 1 << 5
-  };
+/* Macros for character classes; depends on strict-mode  */
+#define CR                  '\r'
+#define LF                  '\n'
+#define LOWER(c)            (unsigned char)(c | 0x20)
+#define TOKEN(c)            (tokens[(unsigned char)c])
+#define IS_ALPHA(c)         ((c) >= 'a' && (c) <= 'z')
+#define IS_NUM(c)           ((c) >= '0' && (c) <= '9')
+#define IS_ALPHANUM(c)      (IS_ALPHA(LOWER(c)) || IS_NUM(c))
 
-
-#define CR '\r'
-#define LF '\n'
-#define LOWER(c) (unsigned char)(c | 0x20)
-#define TOKEN(c) tokens[(unsigned char)c]
+#if HTTP_PARSER_STRICT
+#define IS_URL_CHAR(c)      (normal_url_char[(unsigned char) (c)])
+#define IS_HOST_CHAR(c)     (IS_ALPHANUM(c) || (c) == '.' || (c) == '-')
+#else
+#define IS_URL_CHAR(c)                                                         \
+  (normal_url_char[(unsigned char) (c)] || ((c) & 0x80))
+#define IS_HOST_CHAR(c)                                                        \
+  (IS_ALPHANUM(c) || (c) == '.' || (c) == '-' || (c) == '_')
+#endif
 
 
 #define start_state (parser->type == HTTP_REQUEST ? s_start_req : s_start_res)
@@ -324,6 +333,7 @@ size_t http_parser_execute (http_parser *parser,
                             size_t len)
 {
   char c, ch;
+  int8_t unhex_val;
   const char *p = data, *pe;
   int64_t to_read;
 
@@ -480,7 +490,7 @@ size_t http_parser_execute (http_parser *parser,
           break;
         }
 
-        if (ch < '0' || ch > '9') goto error;
+        if (!IS_NUM(ch)) goto error;
 
         parser->http_major *= 10;
         parser->http_major += ch - '0';
@@ -491,7 +501,7 @@ size_t http_parser_execute (http_parser *parser,
 
       /* first digit of minor HTTP version */
       case s_res_first_http_minor:
-        if (ch < '0' || ch > '9') goto error;
+        if (!IS_NUM(ch)) goto error;
         parser->http_minor = ch - '0';
         state = s_res_http_minor;
         break;
@@ -504,7 +514,7 @@ size_t http_parser_execute (http_parser *parser,
           break;
         }
 
-        if (ch < '0' || ch > '9') goto error;
+        if (!IS_NUM(ch)) goto error;
 
         parser->http_minor *= 10;
         parser->http_minor += ch - '0';
@@ -515,7 +525,7 @@ size_t http_parser_execute (http_parser *parser,
 
       case s_res_first_status_code:
       {
-        if (ch < '0' || ch > '9') {
+        if (!IS_NUM(ch)) {
           if (ch == ' ') {
             break;
           }
@@ -528,7 +538,7 @@ size_t http_parser_execute (http_parser *parser,
 
       case s_res_status_code:
       {
-        if (ch < '0' || ch > '9') {
+        if (!IS_NUM(ch)) {
           switch (ch) {
             case ' ':
               state = s_res_status;
@@ -580,7 +590,7 @@ size_t http_parser_execute (http_parser *parser,
 
         CALLBACK2(message_begin);
 
-        if (ch < 'A' || 'Z' < ch) goto error;
+        if (!IS_ALPHA(LOWER(ch))) goto error;
 
       start_req_method_assign:
         parser->method = (enum http_method) 0;
@@ -594,7 +604,9 @@ size_t http_parser_execute (http_parser *parser,
           case 'M': parser->method = HTTP_MKCOL; /* or MOVE, MKACTIVITY, MERGE, M-SEARCH */ break;
           case 'N': parser->method = HTTP_NOTIFY; break;
           case 'O': parser->method = HTTP_OPTIONS; break;
-          case 'P': parser->method = HTTP_POST; /* or PROPFIND or PROPPATCH or PUT */ break;
+          case 'P': parser->method = HTTP_POST;
+            /* or PROPFIND or PROPPATCH or PUT or PATCH */
+            break;
           case 'R': parser->method = HTTP_REPORT; break;
           case 'S': parser->method = HTTP_SUBSCRIBE; break;
           case 'T': parser->method = HTTP_TRACE; break;
@@ -620,6 +632,8 @@ size_t http_parser_execute (http_parser *parser,
             parser->method = HTTP_CHECKOUT;
           } else if (index == 2  && ch == 'P') {
             parser->method = HTTP_COPY;
+          } else {
+            goto error;
           }
         } else if (parser->method == HTTP_MKCOL) {
           if (index == 1 && ch == 'O') {
@@ -630,11 +644,15 @@ size_t http_parser_execute (http_parser *parser,
             parser->method = HTTP_MSEARCH;
           } else if (index == 2 && ch == 'A') {
             parser->method = HTTP_MKACTIVITY;
+          } else {
+            goto error;
           }
         } else if (index == 1 && parser->method == HTTP_POST && ch == 'R') {
           parser->method = HTTP_PROPFIND; /* or HTTP_PROPPATCH */
         } else if (index == 1 && parser->method == HTTP_POST && ch == 'U') {
           parser->method = HTTP_PUT;
+        } else if (index == 1 && parser->method == HTTP_POST && ch == 'A') {
+          parser->method = HTTP_PATCH;
         } else if (index == 2 && parser->method == HTTP_UNLOCK && ch == 'S') {
           parser->method = HTTP_UNSUBSCRIBE;
         } else if (index == 4 && parser->method == HTTP_PROPFIND && ch == 'P') {
@@ -659,9 +677,13 @@ size_t http_parser_execute (http_parser *parser,
 
         c = LOWER(ch);
 
-        if (c >= 'a' && c <= 'z') {
+        /* Proxied requests are followed by scheme of an absolute URI (alpha).
+         * CONNECT is followed by a hostname, which begins with alphanum.
+         * All other methods are followed by '/' or '*' (handled above).
+         */
+        if (IS_ALPHA(ch) || (parser->method == HTTP_CONNECT && IS_NUM(ch))) {
           MARK(url);
-          state = s_req_schema;
+          state = (parser->method == HTTP_CONNECT) ? s_req_host : s_req_schema;
           break;
         }
 
@@ -672,17 +694,11 @@ size_t http_parser_execute (http_parser *parser,
       {
         c = LOWER(ch);
 
-        if (c >= 'a' && c <= 'z') break;
+        if (IS_ALPHA(c)) break;
 
         if (ch == ':') {
           state = s_req_schema_slash;
           break;
-        } else if (ch == '.') {
-          state = s_req_host;
-          break;
-        } else if ('0' <= ch && ch <= '9') {
-          state = s_req_host;
-          break;
         }
 
         goto error;
@@ -700,9 +716,7 @@ size_t http_parser_execute (http_parser *parser,
 
       case s_req_host:
       {
-        c = LOWER(ch);
-        if (c >= 'a' && c <= 'z') break;
-        if ((ch >= '0' && ch <= '9') || ch == '.' || ch == '-') break;
+        if (IS_HOST_CHAR(ch)) break;
         switch (ch) {
           case ':':
             state = s_req_port;
@@ -719,6 +733,9 @@ size_t http_parser_execute (http_parser *parser,
             CALLBACK(url);
             state = s_req_http_start;
             break;
+          case '?':
+            state = s_req_query_string_start;
+            break;
           default:
             goto error;
         }
@@ -727,7 +744,7 @@ size_t http_parser_execute (http_parser *parser,
 
       case s_req_port:
       {
-        if (ch >= '0' && ch <= '9') break;
+        if (IS_NUM(ch)) break;
         switch (ch) {
           case '/':
             MARK(path);
@@ -741,6 +758,9 @@ size_t http_parser_execute (http_parser *parser,
             CALLBACK(url);
             state = s_req_http_start;
             break;
+          case '?':
+            state = s_req_query_string_start;
+            break;
           default:
             goto error;
         }
@@ -749,7 +769,7 @@ size_t http_parser_execute (http_parser *parser,
 
       case s_req_path:
       {
-        if (normal_url_char[(unsigned char)ch]) break;
+        if (IS_URL_CHAR(ch)) break;
 
         switch (ch) {
           case ' ':
@@ -787,7 +807,7 @@ size_t http_parser_execute (http_parser *parser,
 
       case s_req_query_string_start:
       {
-        if (normal_url_char[(unsigned char)ch]) {
+        if (IS_URL_CHAR(ch)) {
           MARK(query_string);
           state = s_req_query_string;
           break;
@@ -823,7 +843,7 @@ size_t http_parser_execute (http_parser *parser,
 
       case s_req_query_string:
       {
-        if (normal_url_char[(unsigned char)ch]) break;
+        if (IS_URL_CHAR(ch)) break;
 
         switch (ch) {
           case '?':
@@ -860,7 +880,7 @@ size_t http_parser_execute (http_parser *parser,
 
       case s_req_fragment_start:
       {
-        if (normal_url_char[(unsigned char)ch]) {
+        if (IS_URL_CHAR(ch)) {
           MARK(fragment);
           state = s_req_fragment;
           break;
@@ -897,7 +917,7 @@ size_t http_parser_execute (http_parser *parser,
 
       case s_req_fragment:
       {
-        if (normal_url_char[(unsigned char)ch]) break;
+        if (IS_URL_CHAR(ch)) break;
 
         switch (ch) {
           case ' ':
@@ -975,7 +995,7 @@ size_t http_parser_execute (http_parser *parser,
           break;
         }
 
-        if (ch < '0' || ch > '9') goto error;
+        if (!IS_NUM(ch)) goto error;
 
         parser->http_major *= 10;
         parser->http_major += ch - '0';
@@ -986,7 +1006,7 @@ size_t http_parser_execute (http_parser *parser,
 
       /* first digit of minor HTTP version */
       case s_req_first_http_minor:
-        if (ch < '0' || ch > '9') goto error;
+        if (!IS_NUM(ch)) goto error;
         parser->http_minor = ch - '0';
         state = s_req_http_minor;
         break;
@@ -1006,7 +1026,7 @@ size_t http_parser_execute (http_parser *parser,
 
         /* XXX allow spaces after digit? */
 
-        if (ch < '0' || ch > '9') goto error;
+        if (!IS_NUM(ch)) goto error;
 
         parser->http_minor *= 10;
         parser->http_minor += ch - '0';
@@ -1024,6 +1044,7 @@ size_t http_parser_execute (http_parser *parser,
       }
 
       case s_header_field_start:
+      header_field_start:
       {
         if (ch == CR) {
           state = s_headers_almost_done;
@@ -1201,15 +1222,13 @@ size_t http_parser_execute (http_parser *parser,
 
       case s_header_value_start:
       {
-        if (ch == ' ') break;
+        if (ch == ' ' || ch == '\t') break;
 
         MARK(header_value);
 
         state = s_header_value;
         index = 0;
 
-        c = LOWER(ch);
-
         if (ch == CR) {
           CALLBACK(header_value);
           header_state = h_general;
@@ -1223,6 +1242,8 @@ size_t http_parser_execute (http_parser *parser,
           break;
         }
 
+        c = LOWER(ch);
+
         switch (header_state) {
           case h_upgrade:
             parser->flags |= F_UPGRADE;
@@ -1239,7 +1260,7 @@ size_t http_parser_execute (http_parser *parser,
             break;
 
           case h_content_length:
-            if (ch < '0' || ch > '9') goto error;
+            if (!IS_NUM(ch)) goto error;
             parser->content_length = ch - '0';
             break;
 
@@ -1264,7 +1285,6 @@ size_t http_parser_execute (http_parser *parser,
 
       case s_header_value:
       {
-        c = LOWER(ch);
 
         if (ch == CR) {
           CALLBACK(header_value);
@@ -1277,6 +1297,8 @@ size_t http_parser_execute (http_parser *parser,
           goto header_almost_done;
         }
 
+        c = LOWER(ch);
+
         switch (header_state) {
           case h_general:
             break;
@@ -1288,7 +1310,7 @@ size_t http_parser_execute (http_parser *parser,
 
           case h_content_length:
             if (ch == ' ') break;
-            if (ch < '0' || ch > '9') goto error;
+            if (!IS_NUM(ch)) goto error;
             parser->content_length *= 10;
             parser->content_length += ch - '0';
             break;
@@ -1344,7 +1366,7 @@ size_t http_parser_execute (http_parser *parser,
       {
         STRICT_CHECK(ch != LF);
 
-        state = s_header_field_start;
+        state = s_header_value_lws;
 
         switch (header_state) {
           case h_connection_keep_alive:
@@ -1362,6 +1384,18 @@ size_t http_parser_execute (http_parser *parser,
         break;
       }
 
+      case s_header_value_lws:
+      {
+        if (ch == ' ' || ch == '\t')
+          state = s_header_value_start;
+        else
+        {
+          state = s_header_field_start;
+          goto header_field_start;
+        }
+        break;
+      }
+
       case s_headers_almost_done:
       headers_almost_done:
       {
@@ -1463,9 +1497,9 @@ size_t http_parser_execute (http_parser *parser,
         assert(nread == 1);
         assert(parser->flags & F_CHUNKED);
 
-        c = unhex[(unsigned char)ch];
-        if (c == -1) goto error;
-        parser->content_length = c;
+        unhex_val = unhex[(unsigned char)ch];
+        if (unhex_val == -1) goto error;
+        parser->content_length = unhex_val;
         state = s_chunk_size;
         break;
       }
@@ -1479,9 +1513,9 @@ size_t http_parser_execute (http_parser *parser,
           break;
         }
 
-        c = unhex[(unsigned char)ch];
+        unhex_val = unhex[(unsigned char)ch];
 
-        if (c == -1) {
+        if (unhex_val == -1) {
           if (ch == ';' || ch == ' ') {
             state = s_chunk_parameters;
             break;
@@ -1490,7 +1524,7 @@ size_t http_parser_execute (http_parser *parser,
         }
 
         parser->content_length *= 16;
-        parser->content_length += c;
+        parser->content_length += unhex_val;
         break;
       }
 
